import logging
import logging.config
import os

from .runtime_config import RuntimeConfig


def init_logger(verbose: int):
    """Init logger with config from either RuntimeConfig or a default."""
    levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    log_level = levels[verbose if verbose < len(levels) else len(levels) - 1]

    try:
        logging.config.dictConfig(RuntimeConfig().logging)
        logging.debug(
            "Initialized logging from RuntimeConfig(%s)",
            RuntimeConfig(),
        )
    except (KeyError, ValueError) as err:
        logging.basicConfig(level=log_level)
        logging.warning(
            "Could not load logging config (%s), continuing with defaults.",
            err,
        )

    if verbose != 0:
        for name in logging.root.manager.loggerDict:
            logging.getLogger(name).setLevel(log_level)
            RuntimeConfig().logging["loggers"].update(
                {name: {"level": str(logging._levelToName[log_level])}}
            )


def set_and_init_logger(
    verbose: int,
    name: str,
    filename: str,
    level: str = "DEBUG",
    formatter: str = "debug",
    replace: bool = False,
):

    logfile = os.path.join(RuntimeConfig().paths["output_path"], filename)
    num = 0
    while os.path.isfile(logfile):
        num += 1
        logfile = os.path.join(
            RuntimeConfig().paths["output_path"],
            f"{filename.rsplit('.', 1)[0]}_{num}.{filename.rsplit('.', 1)[1]}",
        )
    RuntimeConfig().logging["handlers"][name] = {
        "class": "logging.FileHandler",
        "formatter": formatter,
        "level": level,
        "mode": "w",
        "filename": logfile,
    }
    if replace:
        RuntimeConfig().logging["root"]["handlers"] = [name]
    else:
        RuntimeConfig().logging["root"]["handlers"].append(name)

    init_logger(verbose)



import os
from typing import Any, Dict, List

from ruamel.yaml import YAML

from . import LOG
from .dict_util import update
from .runtime_config import RuntimeConfig


def get_config_files(configs: List[str], default_path: str) -> List[str]:
    """Collection all scenario configuration files.

    The functions looks at the two default location for scenario files,
    i.e., the `config` folder inside the `scenario` module and the
    `midas_scenario` folder defined in the current runtime config.
    Additionally, every custom config provided with the *configs*
    parameter is checked for presence.

    Parameters
    ----------
    configs: list
        A list containing path-like strings pointing to scenario
        configuration files each.

    Returns
    -------
    list:
        A list of absolute paths to all existing configuration files.

    """

    os.makedirs(default_path, exist_ok=True)

    files = [os.path.join(default_path, f) for f in os.listdir(default_path)]
    user_path = os.path.abspath(RuntimeConfig().paths["scenario_path"])
    os.makedirs(os.path.abspath(user_path), exist_ok=True)
    files.extend([os.path.join(user_path, f) for f in os.listdir(user_path)])

    if configs is not None:
        for ccfg in configs:
            if not ccfg.endswith(".yml"):
                ccfg = f"{ccfg}.yml"
            ccfg = os.path.abspath(ccfg)
            if os.path.isfile(ccfg):
                LOG.debug("Adding custom config at '%s'.", ccfg)
                files.append(ccfg)
            else:
                LOG.warning("Did not found config '%s'.", ccfg)

    return files


def load_configs(files: str) -> Dict[str, Dict[str, Any]]:
    """Load the config files with yaml."""

    configs = dict()
    yaml = YAML(typ="safe", pure=True)
    for path in files:
        if not path.endswith(".yml"):
            continue

        LOG.debug("Loading config file %s.", path)
        with open(path, "r") as yaml_file:
            config = yaml.load(yaml_file)

        if not config:
            LOG.error("Scenario file at '%s' is empty. Skipping!")
            continue

        for key in config:
            if key in configs:
                LOG.critical(
                    "Scenario name with key '%s' does already exist. "
                    "Please choose a different key in file '%s'.",
                    key,
                    path,
                )
                raise ValueError(
                    f"Scenario '{key}' in file {path} is duplicated. "
                    "Please choose a different name."
                )
        update(configs, config)

    return configs


def normalize(params: Dict[str, Any]):
    """Apply some auto corrections for the parameter dictionary.

    Corrects, e.g., the end definition '15*60' to 900.
    The corrections will be performed inplace.

    """
    for key, val in params.items():
        # Search recusively
        if isinstance(val, dict):
            normalize(val)

        # Correct multiplications
        if isinstance(val, str):
            if "*" in val:
                parts = val.split("*")
                product = 1
                try:
                    for part in parts:
                        product *= float(part)

                    if key in ["step_size", "end"]:
                        product = int(product)
                    params[key] = product
                    LOG.debug(
                        "Corrected value for key %s (%s -> %f).",
                        key,
                        val,
                        product,
                    )
                except ValueError:
                    # Not a multiplication
                    pass
            if val.lower() == "true":
                val = True
                LOG.debug(
                    "Corrected value for key %s ('true' -> bool(True)).",
                    key,
                )
            if val.lower() == "false":
                val = False
                LOG.debug(
                    "Corrected value for key %s ('false' -> bool(False)).",
                    key,
                )

        # Correct mosaik params address which is a tuple and not a list
        if key == "mosaik_params":
            if "addr" in val:
                if isinstance(val["addr"], list):
                    LOG.debug("Corrected mosaik_params.")
                    val["addr"] = (val["addr"][0], val["addr"][1])



"""This module contains a set of utility functions for dictionaries."""
import collections.abc
from typing import Any, Dict, List, Union

import numpy as np

from . import LOG

Numeric = Union[
    bool,
    int,
    float,
    np.int16,
    np.int32,
    np.int64,
    np.float16,
    np.float32,
    np.float64,
    str,
]


def update(src: Dict[str, Any], upd: Dict[str, Any]) -> Dict[str, Any]:
    """Recursive update of dictionaries.

    See stackoverflow:

        https://stackoverflow.com/questions/3232943/
        update-value-of-a-nested-dictionary-of-varying-depth

    """
    for key, val in upd.items():
        if isinstance(val, collections.abc.Mapping):
            src[key] = update(src.get(key, {}), val)
        else:
            src[key] = val
    return src


def convert(src: Dict[str, Any]) -> Dict[str, Any]:
    """Recursive conversion to basic data types."""
    k2d = list()
    for key, val in src.items():
        nkey = convert_val(key)
        if type(nkey) != type(key):
            k2d.append(key)
        key = nkey
        if isinstance(val, collections.abc.Mapping):
            src[key] = convert(val)
        elif isinstance(val, (list, np.ndarray)):
            src[key] = convert_list(val)
        else:
            src[key] = convert_val(val)
    for key in k2d:
        src.pop(key)
    return src


def convert_list(old_list: List[Any]) -> List[Any]:
    new_list = list()
    for val in old_list:
        if isinstance(val, collections.abc.Mapping):
            new_list.append(convert(val))
        elif isinstance(val, (list, tuple, np.ndarray)):
            new_list.append(convert_list(val))
        else:
            new_list.append(convert_val(val))
    return new_list


def convert_val(val: Numeric) -> Union[bool, int, float, str]:
    if isinstance(val, bool):
        return val
    if isinstance(val, (int, np.int16, np.int32, np.int64)):
        return int(val)
    if isinstance(val, (float, np.float16, np.float32, np.float64)):
        return float(val)

    try:
        return strtobool(val)
    except (ValueError, TypeError, AttributeError):
        pass
    try:
        return int(val)
    except (ValueError, TypeError):
        pass
    try:
        return float(val)
    except (ValueError, TypeError):
        pass
    try:
        return str(val)
    except (ValueError, TypeError):
        LOG.info("Unable to convert value %s", val)

    return "MISSING_VALUE"


def strtobool(val: str) -> bool:
    """Convert a string representation of truth to true or false.

    True values are 'y', 'yes', 't', 'true', 'on', and '1'; false
    values are 'n', 'no', 'f', 'false', 'off', and '0'. Raises
    ValueError if 'val' is anything else.

    This is mostly the implementation of the now deprecated
    distutils package. However, it is change in that regard that an
    actual bool type value is returned instead of 1 or 0.

    """
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(f"Invalid truth value {val}")


def tobool(val: Union[bool, int, float, str]) -> bool:
    """Convert any basic type representation to true or false"""

    if isinstance(val, bool):
        return val

    elif isinstance(val, int):
        return val != 0

    elif isinstance(val, float):
        return val != 0.0

    elif isinstance(val, str):
        return strtobool(val)

    else:
        raise ValueError(
            f"Cannot convert value {val} of type {type(val)} to bool."
        )



"""This module contains the base simulator class for certain data
simulators.

The models itself are simple data providers.
"""
import logging
from datetime import datetime, timedelta, timezone

import mosaik_api
import numpy as np
from midas.util.dict_util import strtobool

from .dateformat import GER

LOG = logging.getLogger(__name__)


class BaseDataSimulator(mosaik_api.Simulator):
    """A mosaik simulator for base data models data.

    This model provides the mosaik API for the :class:`.DataModel`.
    However, this simulator is not expected to be used directly.

    Parameters
    ----------
    meta : dict
        The meta dict this simulator should pass to mosaik_api.

    """

    def __init__(self, meta):
        super().__init__(meta)
        self.sid = None
        self.models = dict()

        self.step_size = None
        self.now_dt = None
        self.seed = None
        self.seed_max = None
        self.rng = None
        self.interpolate = None
        self.randomize_data = None
        self.cos_phi = None
        self.randomize_cos_phi = None

        self._sim_time: int = 0

    def init(self, sid, **sim_params):
        """Called exactly ones after the simulator has been started.

        Parameters
        ----------
        sid : str
            Simulator ID for this simulator.
        step_size : int, optional
            Step size for this simulator. Defaults to 900.
        now_dt : str
            An ISO datestring with Europe/Berlin timezone.
        interpolate : bool
            If *True*, interpolation is applied when step size is lower
            than the resolution of the data.
        randomize_data : bool
            If *True*, a small random noise is added to the data in
            each step.
        cos_phi : float
            A static cos phi used if no values for q are present in the
            data.
        randomize_cos_phi : bool
            If *True*, the cos phi is randomized. Applies only to data
            without values for q.
        seed : int
            A seed for the random number generator. Random numbers are
            only used, if one of the randomize flags is set.

        Returns
        -------
        dict
            The meta dict (set by mosaik_api.Simulator)

        """

        # General params
        self.sid = sid
        if "step_size" not in sim_params:
            LOG.debug(
                "Param *step_size* not provided. "
                "Using default step size of 900."
            )
        self.step_size = int(sim_params.get("step_size", 900))
        self.now_dt = datetime.strptime(
            sim_params["start_date"], GER
        ).astimezone(timezone.utc)
        self.interpolate = sim_params.get("interpolate", False)
        try:
            self.interpolate = bool(strtobool(self.interpolate))
        except AttributeError:
            pass

        self.randomize_data = sim_params.get("randomize_data", False)
        self.cos_phi = sim_params.get("cos_phi", 0.9)
        self.randomize_cos_phi = sim_params.get("randomize_cos_phi", False)

        # RNG
        self.seed = sim_params.get("seed", None)
        self.seed_max = sim_params.get("seed_max", 1_000_000)
        if self.seed is not None:
            self.rng = np.random.RandomState(self.seed)
        else:
            LOG.debug("No seed provided. Using random seed.")
            self.rng = np.random.RandomState()

        return self.meta

    def create(self, num, model, **model_params):
        """Initialize the simulation model instance (entity)

        Parameters
        ----------
        num : int
            The number of models to create.
        model : str
            The name of the models to create. Must be present inside
            the simulator's meta.

        Returns
        -------
        list
            A list with information on the created entity.

        """
        raise NotImplementedError

    def step(self, time, inputs, max_advance=0):
        """Perform a simulation step.

        Parameters
        ----------
        time : int
            The current simulation step (by convention in seconds since
            simulation start.
        inputs : dict
            A *dict* containing inputs for entities of this simulator.

        Returns
        -------
        int
            The next step this simulator wants to be stepped.

        """

        self._sim_time = time
        for model in self.models.values():
            model.cos_phi = self.cos_phi
            model.now_dt = self.now_dt

        for eid, attrs in inputs.items():
            for attr, src_ids in attrs.items():
                if attr == "cos_phi":
                    assert (
                        len(src_ids) == 1
                    ), "Only one provider for cos_phi allowed"
                    self.models[eid].cos_phi = list(src_ids.values())[0]
                else:
                    LOG.info("Detected not supported input %s.", attr)

        for eid, model in self.models.items():
            model.step()

        self.now_dt += timedelta(seconds=self.step_size)

        return time + self.step_size

    def get_data(self, outputs):
        """Return the requested output (if feasible).

        Parameters
        ----------
        outputs : dict
            A *dict* containing requested outputs of each entity.

        Returns
        -------
        dict
            A *dict* containing the values of the requested outputs.

        """
        data = dict()
        for eid, attrs in outputs.items():
            model = eid.split("-")[0]
            data[eid] = dict()
            for attr in attrs:
                if attr == "outbox":
                    data[eid][attr] = {
                        "p_mw": getattr(self.models[eid], "p_mw"),
                        "q_mvar": getattr(self.models[eid], "q_mvar"),
                    }
                elif attr in self.meta["models"][model]["attrs"]:
                    data[eid][attr] = getattr(self.models[eid], attr)
                else:
                    LOG.info("Unknown output attribute %s.", attr)

        return data



"""This module contains the abstract base class for all upgrade
modules. Provides a basic workflow for the definition of new
upgrades.

"""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
import mosaik
import pandas as pd

if TYPE_CHECKING:
    from midas.scenario.scenario import Scenario
    from mosaik.scenario import World


LOG = logging.getLogger(__name__)


class UpgradeModule(ABC):
    """Base class for upgrade modules.

    Parameters
    ----------
    name: str
        The name of this module. This is the name that is referenced
        in the yaml file and must be present in the module load order.
    default_sim_config_name: str
        This is the name that is placed in the mosaik sim config. The
        simulator will be known under this name within mosaik.
        This is a default value; individual simulators may overwrite
        this.
    default_import_str: str
        This is default import path for this module if `python` is used
        to start this simulator.
    default_cmd_str: str
        This is the default command to start this module if `cmd` is
        used to start this simulator.
    default_scope_name: str
        If no scope is provided for this module, a new scope will be
        created with this name.

    """

    def __init__(
        self,
        module_name: str,
        default_scope_name: str,
        default_sim_config_name: str,
        default_import_str: str,
        default_cmd_str: Optional[str] = None,
        log: Optional[logging.Logger] = None,
    ):
        self.module_name: str = module_name
        self.default_scope_name: str = default_scope_name
        self.default_sim_config_name: str = default_sim_config_name
        self.default_import_str: str = default_import_str
        self.default_cmd_str = default_cmd_str
        self.world: "World"
        self.scenario: "Scenario"
        self.params: dict
        self.module_params: Dict[str, Any]
        self.sim_params: dict
        self.sim_key: str = ""
        self.scope_name: str = ""
        self.scopes: dict = {}
        self.logger: logging.Logger

        if log is None:
            self.logger = LOG
        else:
            self.logger = log

    def upgrade(self, scenario: "Scenario", params: dict):
        """Upgrade the scenario with this module.

        Adds the functionality provided by this upgrade to the
        scenario, i.e., define and start a simulator in the mosaik
        world, instantiate models, and add connections to other
        existing models.

        Parameters
        ----------
        scenario : :class:`midas.scenario.scenario.Scenario`
            The scenario containing reference to everything created in
            former upgrades.
        params : dict
            A *dict* containing the content of the config files and
            additional information generated during other upgrades.

        """
        self.world = scenario.world
        self.scenario = scenario
        self.params = params
        self.module_params = self._check_module_params()

        for scope_name, sim_params in self.module_params.items():
            if not isinstance(sim_params, dict):
                continue

            self.scope_name = scope_name
            self.sim_params = sim_params

            self._check_sim_params(self.module_params)
            self.scopes[scope_name] = sim_params

            self._start_simulator()

            self.start_models()

            self.connect()
            if self.sim_params["with_arl"]:
                self.get_sensors()
                self.get_actuators()

            if not self.sim_params["no_db"]:
                try:
                    self.connect_to_db()
                except Exception as err:
                    LOG.error(
                        "Could not connect to the database module: %s "
                        "Did you add it to the modules key of your scenario? "
                        "e.g.: modules: [store, ...]",
                        err,
                    )

            if self.sim_params["with_timesim"]:
                try:
                    self.connect_to_timesim()
                except Exception as err:
                    LOG.error(
                        "Could not connect to the timesim module: %s "
                        "Did you add it to the modules key of your scenario? "
                        "e.g.: modules: [timesim, ...]",
                        err,
                    )

    def _check_module_params(self) -> Dict[str, Any]:
        module_params = self.params.setdefault(
            f"{self.module_name}_params", {}
        )

        if not module_params:
            module_params[self.default_scope_name] = {}

        module_params.setdefault("sim_name", self.default_sim_config_name)
        module_params.setdefault("cmd", self.scenario.base.cmd)
        if module_params["cmd"] == "python":
            module_params.setdefault("import_str", self.default_import_str)
        elif module_params["cmd"] == "cmd":
            module_params.setdefault("import_str", self.default_cmd_str)
        else:
            LOG.error(
                "Invalid or unsupported cmd string: %s", module_params["cmd"]
            )
        module_params.setdefault("step_size", self.scenario.base.step_size)
        module_params.setdefault("no_db", self.scenario.base.no_db)
        module_params.setdefault(
            "with_timesim", self.scenario.base.with_timesim
        )
        module_params.setdefault("no_rng", self.scenario.base.no_rng)
        module_params.setdefault("with_arl", self.scenario.base.with_arl)
        self.check_module_params(module_params)
        return module_params

    @abstractmethod
    def check_module_params(self, module_params: Dict[str, Any]):
        """Is called from within the upgrade method."""
        raise NotImplementedError

    def _check_sim_params(self, module_params: Dict[str, Any]):
        self.sim_params.setdefault("sim_name", module_params["sim_name"])
        self.sim_params.setdefault("cmd", module_params["cmd"])
        if self.sim_params["cmd"] == module_params["cmd"]:
            self.sim_params.setdefault(
                "import_str", module_params["import_str"]
            )
        elif self.sim_params["cmd"] == "python":
            self.sim_params.setdefault("import_str", self.default_import_str)
        else:
            self.sim_params.setdefault("import_str", self.default_cmd_str)

        self.sim_params.setdefault("step_size", module_params["step_size"])
        self.sim_params.setdefault(
            "with_timesim", self.module_params["with_timesim"]
        )
        self.sim_params.setdefault("no_db", module_params["no_db"])
        self.sim_params.setdefault("no_rng", module_params["no_rng"])
        self.sim_params.setdefault("with_arl", module_params["with_arl"])

        self.check_sim_params(module_params)

    @abstractmethod
    def check_sim_params(self, module_params: Dict[str, Any]):
        """Is called from within the upgrade method."""
        raise NotImplementedError

    def _start_simulator(self):
        """Start a certain simulator instance."""

        cmd = self.sim_params.pop("cmd")
        import_str = self.sim_params.pop("import_str")
        with_timesim = self.sim_params.pop("with_timesim")
        no_db = self.sim_params.pop("no_db")
        no_rng = self.sim_params.pop("no_rng")
        with_arl = self.sim_params.pop("with_arl")

        # Place model in the world's *sim_config*
        self.world.sim_config[self.sim_params["sim_name"]] = {cmd: import_str}
        self.scenario.script.simconfig = [
            f"sim_config = {self.world.sim_config}\n"
        ]

        # Create a unique simulator key
        self.sim_key = self.scenario.generate_sim_key(self)

        # Start the simulator if it was not started before
        if not self.scenario.sim_started(self.sim_key):
            self.scenario.add_sim(
                self.sim_key, self.world.start(**self.sim_params)
            )
            self.logger.debug(
                "Started simulator %s (key: %s).",
                getattr(self.scenario.get_sim(self.sim_key), "_sim").sid,
                self.sim_key,
            )
            self.scenario.script.definitions.append(
                f"{self.sim_key}_params = {self.sim_params}\n"
            )
            self.scenario.script.sim_start.append(
                f"{self.sim_key} = world.start(**{self.sim_key}_params)\n"
            )
        self.sim_params["cmd"] = cmd
        self.sim_params["import_str"] = import_str
        self.sim_params["with_timesim"] = with_timesim
        self.sim_params["with_arl"] = with_arl
        self.sim_params["no_db"] = no_db
        self.sim_params["no_rng"] = no_rng

    @abstractmethod
    def start_models(self):
        raise NotImplementedError

    @abstractmethod
    def connect(self):
        raise NotImplementedError

    @abstractmethod
    def connect_to_db(self):
        raise NotImplementedError

    def connect_to_timesim(self):
        pass

    def start_model(
        self, model_key: str, model_name: str, params: Dict[str, Any]
    ) -> str:
        if not self.scenario.model_started(model_key, self.sim_key):
            sim = self.scenario.get_sim(self.sim_key)
            entity = getattr(sim, model_name)(**params)
            self.scenario.add_model(model_key, self.sim_key, entity)

            self.logger.debug(
                "Created model %s (key: %s).",
                entity.full_id,
                model_key,
            )

            self.scenario.script.definitions.append(
                f"{model_key}_params = {params}\n"
            )
            self.scenario.script.model_start.append(
                f"{model_key} = {self.sim_key}.{model_name}(**{model_key}"
                "_params)\n"
            )
            return entity.full_id
        else:
            entity = self.scenario.get_model(model_key, self.sim_key)
            self.logger.info(
                "Model %s (key: %s) already started",
                entity.full_id,
                model_key,
            )
            return entity.full_id

    def connect_entities(
        self,
        from_ent_key: str,
        to_ent_key: str,
        attrs: List[Union[str, Tuple[str, str]]],
        **kwargs,
    ):
        from_entity = self.scenario.get_model(from_ent_key)
        to_entity = self.scenario.get_model(to_ent_key)
        try:
            self.world.connect(from_entity, to_entity, *attrs, **kwargs)
        except mosaik.exceptions.ScenarioError:
            self.logger.exception(
                "Something went wrong with your scenario setup."
            )
        self.logger.debug(
            "Connected %s to %s (%s).",
            from_entity.full_id,
            to_entity.full_id,
            attrs,
        )
        self.scenario.script.connects.append(
            f"world.connect({from_ent_key}, {to_ent_key}, *{attrs}, "
            f"**{kwargs})\n"
        )

    def get_sensors(self):
        pass

    def get_actuators(self):
        pass

    def get_info(self):
        self.logger.debug("I don't provide info!")

    def download(
        self,
        data_path: str,
        tmp_path: str,
        load_if_necessary: bool,
        force: bool,
    ):
        """Download the data set of this module.
        This method should if be overwritten if the module has any data
        sets that need to be downloaded.

        Parameters
        ==========
        data_path: str
            Path to the data folder of MIDAS.
        tmp_path: str
            Path to the temporary folder, where temporary download
            artifacts can be stored. Will be deleted afterwards.
        load_if_necessary: bool
            This flag indicates, if the `load_on_start` key of the
            runtime config should be evaluated to decide, if the module
            should perform a download. If the data set exists, there
            will be no download.
        foree: bool
            This flag allows to force a download and overwrite any
            existing data sets of this module.
        """
        self.logger.debug("I don't provide downloads!")

    def analyze(
        self,
        name: str,
        data: pd.HDFStore,
        output_folder: str,
        start: int,
        end: int,
        step_size: int,
        full: bool,
    ):
        self.logger.debug("I don't provide analyis!")



"""The MIDAS runtime config (adapted from the palaestrAI runtime
config).

"""
from __future__ import annotations

from copy import deepcopy
from io import TextIOWrapper
from os import getcwd
from pathlib import Path
from typing import Any, Dict, TextIO, Union

import appdirs
import ruamel.yaml

from . import LOG

CONFIG_FILE_NAME = "midas-runtime-conf.yml"
CONFIG_FILE_PATHS = [
    f"{appdirs.site_config_dir('midas', 'OFFIS')}/{CONFIG_FILE_NAME}",
    f"{appdirs.user_config_dir('midas', 'OFFIS')}/{CONFIG_FILE_NAME}",
    f"{getcwd()}/{CONFIG_FILE_NAME}",
]
DATA_DIR_NAME = "midas_data"
SCENARIO_DIR_NAME = "midas_scenarios"
OUTPUT_DIR_NAME = "_outputs"
DEFAULT_BASE_CONFIG = {
    "paths": {
        "data_path": (
            f"{appdirs.user_config_dir('midas', 'OFFIS')}/{DATA_DIR_NAME}"
        ),
        # "output_path": f"{getcwd()}/_outputs",
        "output_path": OUTPUT_DIR_NAME,
        "scenario_path": (
            f"{appdirs.user_config_dir('midas', 'OFFIS')}/{SCENARIO_DIR_NAME}"
        ),
    },
    "data": {
        "commercials": [
            {
                "name": "CommercialsRefTMY3.hdf5",
                "load_on_start": True,
                "base_url": "https://openei.org/datasets/files/961/pub/",
                "loc_url": (
                    "COMMERCIAL_LOAD_DATA_E_PLUS_OUTPUT/USA_NY_Rochester-"
                    "Greater.Rochester.Intl.AP.725290_TMY3/RefBldg"
                ),
                "post_fix": "New2004_v1.3_7.1_5A_USA_IL_CHICAGO-OHARE.csv",
                "data_urls": [
                    ["FullServiceRestaurant", "FullServiceRestaurant"],
                    ["Hospital", "Hospital"],
                    ["LargeHotel", "LargeHotel"],
                    ["LargeOffice", "LargeOffice"],
                    ["MediumOffice", "MediumOffice"],
                    ["MidriseApartment", "MidriseApartment"],
                    ["OutPatient", "OutPatient"],
                    ["PrimarySchool", "PrimarySchool"],
                    ["QuickServiceRestaurant", "QuickServiceRestaurant"],
                    ["SecondarySchool", "SecondarySchool"],
                    ["SmallHotel", "SmallHotel"],
                    ["SmallOffice", "SmallOffice"],
                    ["Stand-aloneRetail", "StandaloneRetail"],
                    ["StripMall", "StripMall"],
                    ["SuperMarket", "SuperMarket"],
                    ["Warehouse", "Warehouse"],
                ],
                "el_cols": [
                    "Electricity:Facility [kW](Hourly)",
                    "Fans:Electricity [kW](Hourly)",
                    "Cooling:Electricity [kW](Hourly)",
                    "Heating:Electricity [kW](Hourly)",
                ],
            },
        ],
        "default_load_profiles": [
            {
                "name": "DefaultLoadProfiles.hdf5",
                "load_on_start": True,
                "base_url": "https://www.bdew.de/media/documents/Profile.zip",
                "filename": "Repräsentative Profile VDEW.xls",
                "sheet_names": [
                    "H0",
                    "G0",
                    "G1",
                    "G2",
                    "G3",
                    "G4",
                    "G5",
                    "G6",
                    "L0",
                    "L1",
                    "L2",
                ],
                "seasons": [
                    ["Winter", "winter"],
                    ["Sommer", "summer"],
                    ["Übergangszeit", "transition"],
                ],
                "days": [
                    ["Samstag", "saturday"],
                    ["Sonntag", "sunday"],
                    ["Werktag", "weekday"],
                ],
            }
        ],
        "generator_timeseries": [
            {
                "base_url": "https://ds.50hertz.com/api/",
                "pv_url": "PhotovoltaicActual",
                "wind_url": "WindPowerActual",
                "postfix": "/DownloadFile?fileName=",
                "name": "GenWindPV50Hertz.hdf5",
                "load_on_start": True,
                "first_year": 2019,
                "last_year": 2019,
            }
        ],
        "simbench": [
            {
                "name": "1-LV-rural3--0-sw.hdf5",
                "load_on_start": True,
            }
        ],
        "smart_nord": [
            {"name": "SmartNordProfiles.hdf5", "load_on_start": True}
        ],
        "weather": [
            {
                "name": "WeatherBre2009-2020.hdf5",
                "load_on_start": True,
                "base_url": (
                    "https://opendata.dwd.de/climate_environment/CDC/"
                    "observations_germany/climate/hourly/"
                ),
                "air_url": (
                    "air_temperature/historical/stundenwerte_TU_00691_"
                    "19490101_"
                ),
                "wind_url": (
                    "wind/historical/stundenwerte_FF_00691_19260101_"
                ),
                "sun_url": ("sun/historical/stundenwerte_SD_00691_19510101_"),
                "cloud_url": (
                    "cloudiness/historical/stundenwerte_N_00691_19490101_"
                ),
                "solar_url": "solar/stundenwerte_ST_00691_row.zip",
                "post_fix": "20211231_hist.zip",
                "start_date": "2009-01-01 00:00:00",
            }
        ],
    },
    "modules": {
        "default_modules": [
            ["store", "midas.modules.store.module:DatabaseModule"],
            ["timesim", "midas.modules.timesim.module:TimeSimModule"],
            ["powergrid", "midas.modules.powergrid.module:PowergridModule"],
            ["sbdata", "midas.modules.sbdata.module:SimbenchDataModule"],
            ["sndata", "midas.modules.sndata.module:SmartNordDataModule"],
            ["comdata", "midas.modules.comdata.module:CommercialDataModule"],
            ["dlpdata", "midas.modules.dlpdata.module:DLPDataModule"],
            ["pwdata", "midas.modules.pwdata.module:PVWindDataModule"],
            ["weather", "midas.modules.weather.module:WeatherDataModule"],
            ["der", "pysimmods.mosaik.midas_module:PysimmodsModule"],
            ["goa", "midas.modules.goa.module:GridOperatorModule"],
        ],
        "custom_modules": [],
    },
    "misc": {
        "seed_max": 1_000_000,
        "use_pprint": False,
        "key_value_logs": False,
    },
    "logging": {
        "version": 1,
        "formatters": {
            "simple": {
                "format": ("%(asctime)s [%(name)s][%(levelname)s] %(message)s")
            },
            "debug": {
                "format": (
                    "%(asctime)s [%(name)s (%(process)d)][%(levelname)s] "
                    "%(message)s (%(module)s.%(funcName)s in %(filename)s"
                    ":%(lineno)d)"
                )
            },
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "level": "INFO",
                "formatter": "simple",
                "stream": "ext://sys.stdout",
            },
            "console-debug": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "debug",
                "stream": "ext://sys.stdout",
            },
            "logfile": {
                "class": "logging.FileHandler",
                "level": "INFO",
                "formatter": "simple",
                "filename": "midas.log",
                "mode": "w",
            },
            "logfile-debug": {
                "class": "logging.FileHandler",
                "level": "DEBUG",
                "formatter": "debug",
                "filename": "midas_debug.log",
                "mode": "w",
            },
        },
        "loggers": {
            "midas.api": {"level": "WARNING"},
            "midas.modules": {"level": "WARNING"},
            "midas.scenario": {"level": "WARNING"},
            "midas.tools": {"level": "WARNING"},
            "midas.util": {"level": "WARNING"},
            "pysimmods": {"level": "WARNING"},
        },
        "root": {
            "level": "ERROR",
            "handlers": [
                "console",
                "logfile-debug",
            ],
        },
    },
}


class _RuntimeConfig:
    """Application-wide runtime configuration.

    This singleton class provides an application-wide runtime configuration
    and transparently hides all sources from the rest of the application.
    """

    DEFAULT_CONFIG = deepcopy(DEFAULT_BASE_CONFIG)

    _instance = None

    def __init__(self):
        self._config_file_path = None
        self.config_search_path = CONFIG_FILE_PATHS

        # The loaded configuration is what RuntimeConfig.load gave us.
        # It remeains immutable after loading.
        self._loaded_configuration = dict()

    def _get(self, key: str, default=None, exception=None) -> Any:
        """Retrieves a config key.

        Retrieves any config key; if not set, it queries the config
        dictionary; if it isn't present there, it returns the given
        default value. It also sets the value in the current object
        as side-effect.

        """
        lkey = f"_{key}"
        if lkey not in self.__dict__:
            try:
                self.__dict__[lkey] = self._loaded_configuration[key]
            except KeyError:
                if default:
                    self.__dict__[lkey] = default
                else:
                    self.__dict__[lkey] = _RuntimeConfig.DEFAULT_CONFIG[key]
                if exception:
                    raise KeyError(exception)
        return self.__dict__[lkey]

    def reset(self):
        """Resets the runtime configuration to empty state."""
        for key in list(self._loaded_configuration.keys()) + list(
            _RuntimeConfig.DEFAULT_CONFIG.keys()
        ):
            try:
                del self.__dict__[f"_{key}"]
            except KeyError:
                pass
        self._loaded_configuration = dict()
        self._config_file_path = None

    @property
    def logging(self) -> Dict:
        """Configuration of all subsystem loggers.

        Returns
        -------
        dict
            A logging configuration that can be fed into
            `logging.DictConfig`.

        """
        return self._get(
            "logging", exception="Sorry, no logging config in the config file."
        )

    # TODO: Add more properties
    @property
    def paths(self) -> Dict:
        """Return the configured paths."""
        return self._get("paths", exception="Sorry, not paths configured.")

    @property
    def data(self) -> Dict:
        """Return the configured datasets."""
        data = self._get("data", exception="Sorry, no datasets configured.")
        for key, values in DEFAULT_BASE_CONFIG["data"].items():
            data.setdefault(key, values)
        return data

    @property
    def modules(self) -> Dict:
        return self._get("modules", exception="Sorry, no modules configured.")

    @property
    def misc(self) -> Dict:
        """Return miscellaneous options."""
        return self._get(
            "misc", exception="Sorry, no miscellaneous options configured."
        )

    def load(
        self, stream_or_dict: Union[dict, TextIO, str, Path, None] = None
    ):
        """Load the configuration from an external source.

        The runtime configuration is initialized from the default
        configuration in ::`_RuntimeConfig.DEFAULT_CONFIG`. This method
        then iterates through the list in
        ::`_RuntimeConfig.CONFIG_FILE_PATHS`, subsequently updating the
        existing configuration with new values found. Finally, the
        given ::`stream_or_dict` parameter is used if present,
        ultimately taking preference over all other values.

        That means that each config file can contain only a portion of
        the overall configuration; it gets updated subsequently.

        Parameters
        ----------
        stream_or_dict: Union[dict, TextIO, str, Path, None]
            Loads the runtime configuration directly from a dictionary
            or as YAML-encoded stream. If no stream is given, the
            default files in ::`.CONFIG_FILE_PATHS` will be tried as
            described.

        """
        if not isinstance(self._loaded_configuration, dict):
            self._loaded_configuration = dict()
        if stream_or_dict is None and self._loaded_configuration:
            # if not stream_or_dict and len(self._loaded_configuration) > 0:
            # Don't load a default config if we already have something;
            # use reset() instead.
            return

        # yml = ruamel.yaml.YAML(typ="safe")
        has_seen_nondefault_config = False
        self._loaded_configuration.update(_RuntimeConfig.DEFAULT_CONFIG)
        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        for fil in CONFIG_FILE_PATHS:
            try:
                LOG.debug("Trying to open configuration file: '%s'", fil)
                with open(fil, "r") as fp:
                    deserialized = yaml.load(fp)
                    if not isinstance(deserialized, dict):
                        LOG.warning(
                            "The contents of '%s' could not be deserialized"
                            " to dict, skipping it.",
                            fil,
                        )
                        continue
                    self._loaded_configuration.update(deserialized)
                    self._config_file_path = fil
                    has_seen_nondefault_config = True
            except IOError:
                continue

        if isinstance(stream_or_dict, dict):
            self._loaded_configuration.update(stream_or_dict)
            self._config_file_path = "(dict)"
            return

        if isinstance(stream_or_dict, str):
            stream_or_dict = Path(stream_or_dict)
        if isinstance(stream_or_dict, Path):
            try:
                stream_or_dict = open(stream_or_dict, "r")
            except OSError:
                LOG.warning(
                    "Failed to load runtime configuration from file at"
                    " '%s', ignoring",
                    stream_or_dict,
                )
        if stream_or_dict is not None:
            try:
                deserialized = yaml.load(stream_or_dict)
                if not isinstance(deserialized, dict):
                    raise TypeError
                self._loaded_configuration.update(deserialized)
                try:
                    self._config_file_path = stream_or_dict.name
                except AttributeError:
                    self._config_file_path = str(stream_or_dict)
                has_seen_nondefault_config = True
            except TypeError:
                LOG.warning(
                    "Failed to load runtime configuration from stream"
                    " at '%s', ignoring it",
                    repr(stream_or_dict),
                )
            finally:
                if isinstance(stream_or_dict, TextIOWrapper):
                    stream_or_dict.close()

        if not has_seen_nondefault_config:
            LOG.info(
                "No runtime configuration given, loaded built-in default."
            )
            self._config_file_path = "(DEFAULT)"

    def to_dict(self) -> Dict:
        return {key: self._get(key) for key in _RuntimeConfig.DEFAULT_CONFIG}

    def __str__(self):
        return f"<RuntimeConfig id=0x{id(self)}> at {self._config_file_path}"

    def __repr__(self):
        return str(self.to_dict())


def RuntimeConfig():
    if _RuntimeConfig._instance is None:
        _RuntimeConfig._instance = _RuntimeConfig()
        try:
            _RuntimeConfig._instance.load()
        except FileNotFoundError:
            from copy import deepcopy

            _RuntimeConfig._instance._loaded_configuration = deepcopy(
                _RuntimeConfig.DEFAULT_CONFIG
            )
    return _RuntimeConfig._instance



"""This module contains the model for data simulators."""

from datetime import timedelta

import numpy as np
from midas.util.compute_q import compute_q


class ForecastDataModel:
    """A model for a forecast of a single load or sgen time series.

    This model can be used to "simulate" the forecast of a time series
    of active and, if provided, reactive power values. The model is
    designed to simulate one year but the time resolution of the data needs
    to be specified. This time resolution can be different than the step
    size for one step but if the step size is lower, values will be
    returned repeatedly.

    If that behavior is not desired, the linear interpolation function
    of the model can be used (not yet supported).This can also be combined
    with the randomization function.
    Those can be activated by passing the corresponding flags (see Parameter section).

    If the data does not contain a reactive power time series, the
    model will calculate reactive power based on the cos phi, which
    should be provided before each step. There is an option to
    randomize the output of the cos phi calculation as well.

    The main input for the model is a datetime object for the time to
    be simulated. Month, day, hour, minute, and second will be used to
    determine the corresponding value of the data set. Year information
    will be dropped.

    This has two consequences. First, the model can be used to simulate
    several years and, secondly, the data set needs to be exactly for
    one year. Larger data sets will be "cropped" to one year and
    smaller data sets will probably lead to an error.

    Parameters
    ----------
    data_p : pandas.DataFrame
        Contains values (either load or sgen) for active power. The
        index columns are simple *int*.
    data_q : pandas.DataFrame
        Contains values (either load or sgen) for reactive power. The
        index columns are simple *int*. If *None* is provided for
        *data_q*, the cos phi is used in each step to calculate a value
        for *q*.
    data_step_size: int
        Timely resolution of the data in seconds.
    scaling : float
        An overall scaling for this model. This scaling is applied in
        each to step both of *p* and *q*.
    seed : int, optional
        A seed for the random number generator.
    interpolate : bool, optional (not yet supported)
        If set to *True*, interpolation is applied when values between
        full 15-minute intervalls are requested.
    randomize_data : bool, optional
        If set to *True*, a normally distributed random noise is
        applied to all outputs.
    randomize_cos_phi : bool, optional (not yet supported)
        If set to *True* and data_q is not provided, the cos phi for
        the calculation of *q* is randomized.
    horizon_hours: float, optional
        Defines for how much into the future, the forecast is given.
        Default: 1 hour.
    date_index : bool, optional
        Set this to *True* if the data has datetime as index instead of
        ints (planned but not yet supported).
    noise_factor: float, optional
        Set this to increase or lower the noise if randomization is
        activated.

    Attributes
    ----------
    now_dt : datetime.datetime
        *now_dt is an input and needs to be provided in each step.*
        The current simulated time. Is used to calculate the index for
        the current value in the time series.
    cos_phi : float
        *cos_phi is an input and needs to be provided in each step.*
        The phase angle is used to calculate reactive power if no reactive
        power time series is provided.
    p_mw : float
        *p_mw is an output.* The active power for the current step.
    q_mvar : float
        *q_mvar is an output.* The reactive power for the current step.

    """

    def __init__(self, data_p, data_q, data_step_size, scaling, **params):
        self.data_p = data_p
        self.data_q = data_q
        self.sps = data_step_size
        self.scaling = scaling
        horizon_hours = params.get("forecast_horizon_hours", 1.0)
        self.horizon_hours = int(horizon_hours)
        self.horizon_minutes = int(60 * (horizon_hours - self.horizon_hours))
        self.horizon_seconds = int(
            3600
            * (horizon_hours - self.horizon_hours - self.horizon_minutes / 60)
        )
        self.date_index = params.get("date_index", False)

        # RNG
        self.seed = params.get("seed", None)
        self.rng = np.random.RandomState(self.seed)

        self.interpolate = params.get("interpolate", False)
        self.randomize_data = params.get("randomize_data", False)
        self.randomize_cos_phi = params.get("randomize_cos_phi", False)
        self.noise_factor = params.get("noise_factor", 0.2)
        self.p_std = self.data_p.std() * self.scaling
        self.p_mw_forecasth_per_a = (
            self.data_p.sum() * self.scaling / self.sps * 3_600
        )
        if self.data_q is not None:
            self.q_std = self.data_q.std() * self.scaling
        else:
            self.q_std = None

        # Inputs
        self.now_dt = None
        self.cos_phi = None

        # Outputs
        self.p_mw_forecast = None
        self.q_mvar_forecast = None

    def step(self):
        """Perform a simulation step."""

        self.p_mw_forecast = []
        self.q_mvar_forecast = []

        self._interpolation()
        self._randomization()

        self._random_cos_phi()
        q_mvar_from_cos_phi = False

        for tidx in range(len(self.p_mw_forecast)):
            if len(self.q_mvar_forecast) == 0 or q_mvar_from_cos_phi:
                q_mvar_from_cos_phi = True
                self.q_mvar_forecast.append(
                    compute_q(self.p_mw_forecast[tidx], self.cos_phi)
                )

            # Numpy multiply failes for int and float if *= is used
            self.p_mw_forecast[tidx] = self.p_mw_forecast[tidx] * self.scaling
            self.q_mvar_forecast[tidx] = (
                self.q_mvar_forecast[tidx] * self.scaling
            )

            # tmp = self.p_mw_forecast[tidx]**2 + self.q_mvar_forecast[tidx]**2
            # if tmp != 0:
            #     self.cos_phi = self.p_mw_forecast[tidx] / math.sqrt(tmp)
            # else:
            #     self.cos_phi = 0

    def _interpolation(self):
        # index from now_dt
        tidx_start = self._time_index_from_dt(self.now_dt)
        # add forecast horizon to now_dt
        end_dt = self.now_dt + timedelta(
            hours=self.horizon_hours,
            minutes=self.horizon_minutes,
            seconds=self.horizon_seconds,
        )
        # index from end_dt
        tidx_end = self._time_index_from_dt(end_dt)

        # Apply interpolation
        if self.interpolate:
            raise NotImplementedError
        else:
            # get values from self.data_p/.data_q
            # within yearly data
            if tidx_start < tidx_end:
                self.p_mw_forecast = list(
                    self.data_p.values[tidx_start:tidx_end]
                )
                if self.data_q is not None:
                    self.q_mvar_forecast = list(
                        self.data_q.values[tidx_start:tidx_end]
                    )
            # across end/start of yearly data
            else:
                p_mw1 = list(self.data_p.values[tidx_start:])
                p_mw2 = list(self.data_p.values[:tidx_end])
                self.p_mw_forecast = p_mw1 + p_mw2
                if self.data_q is not None:
                    q_mvar1 = list(self.data_p.values[tidx_start:])
                    q_mvar2 = list(self.data_p.values[:tidx_end])
                    self.q_mvar_forecast = q_mvar1 + q_mvar2

    def _time_index_from_dt(self, dt):
        # We assume that the dataset starts on the first of the year.
        dif = dt - dt.replace(month=1, day=1, hour=0, minute=0, second=0)
        dif_s = dif.total_seconds()

        tidx = int(dif_s // self.sps) % len(
            self.data_p.index
        )  # step size of the data
        return tidx

    def _interpolate(self, tidx, data):

        raise NotImplementedError

    def _randomization(self):
        # Not yet debugged
        if self.randomize_data:
            for tidx in range(len(self.p_mw_forecast)):
                noise = self.rng.normal(
                    scale=(self.p_std * self.noise_factor), loc=0.0
                )
                self.p_mw_forecast[tidx] = max(
                    0, self.p_mw_forecast[tidx] + noise
                )

                if self.q_mvar_forecast is not None:
                    noise = self.rng.normal(
                        scale=(self.q_std * self.noise_factor), loc=0.0
                    )
                    self.q_mvar_forecast[tidx] = max(
                        0, self.q_mvar_forecast[tidx] + noise
                    )

    def _random_cos_phi(self):
        if len(self.q_mvar_forecast) == 0 and self.randomize_cos_phi:
            self.cos_phi = max(
                0, min(1.0, self.rng.normal(scale=0.02, loc=0.9))
            )
